import sys
import glob
import os.path
import subprocess
import ide._board_specs as _board_specs


import viper.toolbelt as tb
import viper.viperconf as vconf

def addBoardFun(cls):
    def addBoardClos(bls):
        if cls.is_present():
            bls.append(cls)
    return addBoardClos

class Board():
    board_specs = {}

    @classmethod
    def is_present(cls):
        info = tb.get_board_info(cls.board)
        if info:
            return os.path.exists(os.path.join(info["path"],"port","port.def"))
        return False
    @classmethod
    def discover(cls,devices):
        res = []
        for uid in devices:
            dev = devices[uid]
            if cls.devmatcher(dev):
                bb = cls(dev)
                #check if initialization was ok
                if bb.ok:
                    res.append(cls(dev))
        return res

    @classmethod
    def has_vid_pid(cls, dev):
        return dev["vendor"] in cls.ids_vendor and dev["model"] in cls.ids_model

    def __init__(self,device):
        self.port = device["port"] if "port" in device else ""
        self.disk = device["disk"] if "disk" in device else ""
        self.vendor = device["vendor"]
        self.model = device["model"]
        self.uid = device["uid"]
        self.id_serial = device["devstring"]
        self.cmdr = tb
        self.plt = vconf.platform
        self.addr = None
        self.info = tb.get_board_info(self.board)
        if hasattr(self,"preferred_baud"):
            self.baud=self.preferred_baud
        else:
            self.baud = 38400
        self.generate_port()



    def burn(self,bin):
        return "Not burnable!"

    def reset(self):
        pass

    def runcmd(self,tool,*args,cwd=None,shell=False):
        return self.cmdr.run(tool,*args,cwd=cwd,shell=shell)
    
    def generate_port(self):
        if self.board not in Board.board_specs:
            try:
                defines, peripherals, pinout = _board_specs.get_board_specs(self.board,self.info["path"])
                names = set()
                for k,v in pinout.items():
                    names.add(k)
                    for z,w in v.items():
                        names.add(w)
                names.update(peripherals)
                Board.board_specs[self.board] = (defines,peripherals,pinout,names)
            except Exception as e:
                raise e
                self.ok = False
                return
        self.ok=True
        self.defines = Board.board_specs[self.board][0]
        self.peripherals=Board.board_specs[self.board][1]
        self.pinmap = Board.board_specs[self.board][2]
        self.names = Board.board_specs[self.board][3]

    # def generate_names(self):
    #     self.names = set()
    #     for k,v in self.pinmap.items():
    #         self.names.add(k)
    #         for z,w in v.items():
    #             self.names.add(w)
    #     self.names.update(self.peripherals)


    

    # def __repr__(self):
    #     return str(self.toTrimmedDict())

    # def getTag(self):
    #     return self.vendor+":"+self.model

    # def toTrimmedDict(self):
    #     return {
    #         "port":self.port,
    #         "name":self.desc,
    #         "tag":self.getTag(),
    #         "vendor":self.vendor,
    #         "model":self.model,
    #         "id":self.id_serial,
    #         "board":self.board,
    #         "need_restart": self.manual_reset
    #     }

    # def reset(self):
    #     vscr = VDevScript(self,"reset")
    #     try:
    #         return vscr.doit()
    #     finally:
    #         vscr.cleanup()
        
    # def burn(self,binfile):
    #     print(">> Burning",self.desc)
    #     vscr = VDevScript(self,"burn",binfile)
    #     try:
    #         return vscr.doit()
    #     finally:
    #         vscr.cleanup()

    #     