import os
import os.path
import sys
import re


def get_board_specs(boardname,path):
    portfile = os.path.join(path,"port","port.def")

    with open(portfile,"r") as ff:
        lines = ff.readlines()

    mth_list = re.compile('////(.*): (.*)')
    mth_pin = re.compile('\s*/\*\s*([DA0-9]*)\s.*\*/\s*MAKE_PIN\(')
    mth_cls = re.compile('.*MAKE_PIN_CLASS\(([0-9]*),')
    mth_header = re.compile('.*\sconst\s*_(.*)class\[\]\s*STORED')


    names = {
    "SPI":["MOSI","MISO","SCLK"],
    "I2C":["SDA","SCL"],
    "SER":["RX","TX"],
    "CAN":["CANRX","CANTX"],
    "PWM":["PWM"],
    "ICU":["ICU"],
    "ADC":["A"],
    "DAC":["DAC"],
    }


    vpins = {}
    vnames = {}
    vlayout = []
    vprph = []
    cdefs = []

    cfun = None
    clsc = 0
    npin = 0
    for line in lines:
        mth = mth_list.match(line)
        if mth:
            if mth.group(1)=="LAYOUT":
                vlayout = mth.group(2).strip().split(" ")
            elif mth.group(1)=="PERIPHERALS":
                vprph = mth.group(2).strip().split(" ")
            elif mth.group(1)=="CDEFINES":
                cdefs.extend(mth.group(2).strip().split(" "))
            continue

        mth = mth_pin.match(line)
        if mth:
            #print("matched p",line)
            pname = mth.group(1)
            pidx = npin
            if pidx not in vpins:
                vpins[pidx] = {
                    "name":pname,
                    "idx":pidx,
                    "fx":{}
                }
            npin+=1
        else:
            mth = mth_header.match(line)
            if mth:
                #print("matched h",line)
                cfun = mth.group(1).upper()
                clsc = 0
            else:
                mth = mth_cls.match(line)
                if mth:
                    #print("matched c",line)
                    pidx = int(mth.group(1))
                    pin = vpins[pidx]
                    if cfun=="DIGITAL":
                        continue
                    if cfun=="ANALOG":
                        cfun="ADC"

                    if cfun in ["LED","BTN"]:
                        vnames[cfun+str(clsc)]=pidx
                    elif cfun not in pin["fx"]:
                        lst = names[cfun]
                        if len(lst)>1:
                            pin["fx"][lst[clsc%len(lst)]]=lst[clsc%len(lst)]+str(clsc//len(lst))
                        else:
                            pin["fx"][cfun]=lst[0]+str(clsc)
                    clsc+=1

    import pprint
    import collections

    pp = pprint.PrettyPrinter(indent=4)


    pinout = collections.OrderedDict()
    lpins = [v for k,v in vpins.items()]
    lpins.sort(key= lambda x: x["idx"])

    for lpin in lpins:
        pinout[lpin["name"]]=lpin["fx"]

    for vname in vnames:
        pinout[vname]=vpins[vnames[vname]]["fx"]

    #pp.pprint(pinout)
    #pp.pprint(vlayout)
    #pp.pprint(vprph)
    ##pp.pprint(cdefs)

    defines = {
        "BOARD": boardname,
        "LAYOUT": vlayout,
        "CDEFS": cdefs
    }

    return (defines,vprph,pinout)




