// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: http://codemirror.net/LICENSE

(function(mod) {
  if (typeof exports == "object" && typeof module == "object") // CommonJS
    mod(require("../../lib/codemirror"));
  else if (typeof define == "function" && define.amd) // AMD
    define(["../../lib/codemirror"], mod);
  else // Plain browser env
    mod(CodeMirror);
})(function(CodeMirror) {
  "use strict";

  function forEach(arr, f) {
    for (var i = 0, e = arr.length; i < e; ++i) f(arr[i]);
  }

  function arrayContains(arr, item) {
    if (!Array.prototype.indexOf) {
      var i = arr.length;
      while (i--) {
        if (arr[i] === item) {
          return true;
        }
      }
      return false;
    }
    return arr.indexOf(item) != -1;
  }

  function getModFrom(prefix,is_from){
    var found=[]
    for (var i=0;i<Vpm.i_modules.length;i++){
      if (is_from && !Vpm.i_modules[i].from)
        continue
      if (!is_from && Vpm.i_modules[i].from)
        continue
      if (Vpm.i_modules[i].from.indexOf(prefix)===0){
        for(var j=0;j<Vpm.i_modules[i]["import"].length;j++){
          if (is_from){
            found.push(Vpm.i_modules[i].from+" import "+Vpm.i_modules[i]["import"][j])
          } else {
            found.push(Vpm.i_modules[i]["import"][j])
          }
        }
      }
    }
    return found
  }

  function scriptHint(editor, _keywords, getToken) {
    // Find the token at the cursor
    var cur = editor.getCursor(), token = getToken(editor, cur), tprop = token;
    // If it's not a 'word-style' token, ignore the token.

    if (!/^[\w$_]*$/.test(token.string)) {
        token = tprop = {start: cur.ch, end: cur.ch, string: "", state: token.state,
                         className: token.string == ":" ? "python-type" : null};
    }

    if (!context) var context = [];
    context.push(tprop);

    var pword = editor.findWordAt({line:cur.line,ch:cur.ch-token.string.length-2})
    if (pword && pword.anchor && pword.head)
      pword = editor.getRange(pword.anchor,pword.head)
    if (pword=="from"){
      return {
        list: getModFrom(token.string,true),
        from: CodeMirror.Pos(cur.line, token.start),
        to: CodeMirror.Pos(cur.line, token.end),
        no_more: true
      }
    } else if (pword=="import") {
      return {
        list: getModFrom(token.string,false),
        from: CodeMirror.Pos(cur.line, token.start),
        to: CodeMirror.Pos(cur.line, token.end),
        no_more: true
      }
    } else {
      var completionList = getCompletions(token, context);
      completionList = completionList.sort();

      return {list: completionList,
              from: CodeMirror.Pos(cur.line, token.start),
              to: CodeMirror.Pos(cur.line, token.end),
              no_more:false};
    }
  }

  function pythonHint(editor) {
    return scriptHint(editor, pythonKeywordsL, function (e, cur) {return e.getTokenAt(cur);});
  }
  CodeMirror.registerHelper("hint", "python", pythonHint);

  var pythonKeywords = "and del from not while as elif global or with else if pass "
+ "break except import class in raise continue finally is return def for try";
  var pythonKeywordsL = pythonKeywords.split(" ");

  var pythonBuiltins = "abs all enumerate int ord str "
+ "any isinstance sum max min "
+ "tuple len range type "
+ "bytearray bytes float list isinstance "
+ "chr getattr dict "
+ "reversed hasattr "
+ "delattr setattr "
+ "hex print "
+ "pinMode digitalWrite digitalRead thread sleep random onPinRise onPinFall";
  var pythonBuiltinsL = pythonBuiltins.split(" ")

  var viperNames = "__init__ __str__ count split strip index find startswith endswith replace join "
  + "append insert extend pop remove reverse clear copy "
  + "Exception ArithmeticException ZeroDivisionError FloatingPointError OverflowError LookupError "
  + "IndexError KeyError NameError RuntimeError NotImplementedError TypeError ValueError AttributeError UnsupportedError StopIteration "
  + "__iter__ __next__ __getitem__ __setitem__ __delitem__ __len__ __contains__ __add__ __iadd__ __mul__ __imul__ "
  + "get keys items values update popitem upper lower format "
  + "difference intersection isdisjoint issubset issuperset symmetric_difference union "
  + "intersection_update difference_update symmetric_difference_update add discard "
  + "__ctl__ IOError ConnectionError ConnectionAbortedError ConnectionResetError ConnectionRefusedError TimeoutError"

  var viperNamesL = viperNames.split(" ")


  var viperDefines = "INPUT OUTPUT INPUT_PULLUP INPUT_PULLDOWN INPUT_ANALOG OUTPUT_PUSHPULL OUTPUT_OPENDRAIN "
  +"LOW HIGH LOW_TO_HIGH HIGH_TO_LOW "
  +"FALLING_EDGE RISING_EDGE BOTH_EDGES STRIP_LEFT STRIP_RIGHT STRIP_BOTH "
  +"MICROS SECONDS NANOS MILLIS "

  var viperDefinesU = viperDefines.split(" ")
  
  function getCompletions(token, context) {
    var found = [], start = token.string;
    function maybeAdd(str) {
      if (str.lastIndexOf(start, 0) == 0 && !arrayContains(found, str)) found.push(str);
    }

    function gatherCompletions(_obj) {
          forEach(pythonBuiltinsL, maybeAdd);
          forEach(pythonKeywordsL, maybeAdd);
          forEach(viperNamesL, maybeAdd);
          forEach(viperDefinesU, maybeAdd);
    }

    //console.log("VIPERHINT")
    //console.log(token)
    if (context) {
      // If this is a property, see if it belongs to some object we can
      // find in the current environment.
      var obj = context.pop(), base;

      if (obj.type == "variable")
          base = obj.string;
      else if(obj.type == "variable-3")
          base = ":" + obj.string;

      while (base != null && context.length)
        base = base[context.pop().string];
      if (base != null) gatherCompletions(base);
    }
    return found;
  }
});
