var CodeEditor = {
    openTabs: [],
    currentTab: -1,
    getTab: function(pid,file){
        for(var i=0;i<CodeEditor.openTabs.length;i++){
            var tab = CodeEditor.openTabs[i]
            if(tab.pid==pid && tab.file==file){
                tab.pos=i
                return tab;
            }
        }
        return null;
    },
    jumpToFileInEditorCurrent: function(file,line) {
        CodeEditor.editFileInEditor(appData.currentProject,file)
        var pid = appData.currentProject.id
        var tab = CodeEditor.getTab(pid,file)
        CodeEditor.selectTab(tab.pos)
        editors.code.setCursor(line)
    },
    editFileInEditor: function(proj, file) {
        console.log("Edit in editor "+file)
        if (proj.files.hasOwnProperty(file)) {
            var tab = CodeEditor.getTab(proj.id,file)
            if (tab) {
                CodeEditor.selectTab(tab.pos)
                return
            }  
            var m, mode, spec,info;
            if (m = /.+\.([^.]+)$/.exec(file)) {
                info = CodeMirror.findModeByExtension(m[1]);
                if (info) {
                    mode = info.mode;
                    spec = info.mime;
                }
            }
            console.log("mode found is:"+mode)
            tab = {
                pid: proj.id,
                file: file,
                doc: CodeMirror.Doc(window.atob(proj.files[file]),mode)
            }
            CodeEditor.openTabs.push(tab)
            editors.code.swapDoc(tab.doc)
            CodeEditor.currentTab=CodeEditor.openTabs.length-1
        }
        CodeEditor.drawTabs();
    },
    orderTabs: function(){
        if(!appData.currentProject) return
        var tabs = CodeEditor.openTabs
        var plist = []
        var tlist = []
        var newcur = -1;
        for (var i=0;i<tabs.length;i++){
            if(tabs[i].pid==appData.currentProject.id){
                if(CodeEditor.currentTab==i)
                    newcur=plist.length
                plist.push(tabs[i])
            } else {
                if(CodeEditor.currentTab==i)
                    newcur=i
                tlist.push(tabs[i])
            }
        }
        CodeEditor.openTabs=plist.concat(tlist)
        CodeEditor.currentTab=newcur
    }
    ,
    drawTabs: function(){
        var tabstrip = $("#tabstrip")
        tabstrip.empty()
        var pids = {}
        CodeEditor.orderTabs()
        $.each(CodeEditor.openTabs,function(i,tab){
            var idx = Project.get_opened_style(tab.pid)
            pids[tab.pid]=1
            var html = '<div class="tabFileOpen '+((i==CodeEditor.currentTab) ? ("tabSelected"):(""))+' tabStyle'+idx+' '+((tab.pid!=appData.currentProject.id) ? ("readOnlyTab"):("tabCurrent"))+'" '+
                       ' onclick="CodeEditor.selectTab('+i+')" id="tab'+i+'">'+
                       /*'<div class="triangle"></div>'+*/
                       '<span class="filename noselect">'+tab.file+'</span>'+
                       '<div class="tabicons">'+
                       '<span onclick="CodeEditor.closeTab('+i+')" class="fa fa-close noselect" aria-hidden="true"></span>'+
                       '<span onclick="CodeEditor.closeTab('+i+')" class="fa fa-lock noselect" aria-hidden="true"></span>'+
                       '</div>'+
                       '</div>';
            tabstrip.append(html)
        })
        pids = Object.keys(pids)
        Projects.set_as_opened(pids)
        CodeEditor.fill_opened_projects()
        if(CodeEditor.currentTab>=0)
            CodeEditor.selectTab(CodeEditor.currentTab)
        if(CodeEditor.openTabs.length<=0){
            editors.code.setOption("readOnly",true)
            editors.code.swapDoc(CodeMirror.Doc(""))
        }
    },
    saveTab:function(tabidx){
        var tab = CodeEditor.openTabs[tabidx]
        console.log("save Tab "+tabidx)
        console.log(tab)
        if(tab.pid==appData.currentProject.id){
            appData.currentProject.files[tab.file]=window.btoa(tab.doc.getValue())
        }
    },
    closeTab: function(tabidx) {
        CodeEditor.saveTab(tabidx)
        CodeEditor.openTabs.splice(tabidx,1)
        if(CodeEditor.openTabs.length<=0)
            CodeEditor.currentTab=-1
        else if(tabidx>0)
            CodeEditor.currentTab=tabidx-1
        else
            CodeEditor.currentTab=0
        CodeEditor.drawTabs()
        return false
    },
    selectTab: function(tabidx) {
        if(tabidx<0) return
        var curtabidx = CodeEditor.currentTab;
        $("#tab"+curtabidx).removeClass("tabSelected")
        $("#tab"+tabidx).addClass("tabSelected")
        CodeEditor.currentTab=tabidx
        var curtab = CodeEditor.openTabs[tabidx]
        console.log(curtab)
        editors.code.swapDoc(curtab.doc)
        if (curtab.pid==appData.currentProject.id){
            editors.code.setOption("readOnly",false)
        } else {
            editors.code.setOption("readOnly",true)
        }
        CodeEditor.show_info(curtab)
        console.log(curtab.doc)
        return false
    },
    show_info_table:[
        ["W","...","pyl-W-tag","fa-exclamation-circle"],
        ["E","...","pyl-E-tag","fa-exclamation-triangle"],
        ["R","...","pyl-R-tag","fa-circle"],
        ["C","...","pyl-C-tag","fa-circle"],
        ["F","...","pyl-F-tag","fa-exclamation-triangle"],
    ]
    ,
    show_info: function(curtab){
        var cinfo = appData.currentProject.compile_info
        var linfo = appData.currentProject.lint_info
        CodeEditor.clearErrors()

        console.log("Showing info for: "+curtab.file)
        if (linfo && linfo.hasOwnProperty(curtab.file)){
            var info = linfo[curtab.file]
            var table = CodeEditor.show_info_table;
            for (var i=0;i<table.length;i++){
                var T = table[i][0]
                for (var j=0;j<info[T].length;j++){
                    var marker = CodeEditor.makeMarker(info[T][j].type+"\n"+info[T][j].msg,table[i][3],table[i][2])
                    editors.code.setGutterMarker(info[T][j].line - 1, 'py-marker', marker);
                }
                $('.'+table[i][2]).qtip({position: {my: 'top left',at: 'bottom right'},content: {title: "Linter"}});
            }
        }
        if (cinfo && cinfo.hasOwnProperty(curtab.file)){
            var info = cinfo[curtab.file]
            for (var j=0;j<info["E"].length;j++){
                var marker= CodeEditor.makeMarker(info["E"][j].msg,"fa-times-circle","pyc-error-tag")
                editors.code.setGutterMarker(info["E"][j].line - 2, "py-marker", marker);
            }
            $('.pyc-error-tag').qtip({position: {my: 'top left',at: 'bottom right'},content: {title: "Compiler"}});
        }
    }
    ,
    fixcode: function(){
        editors.code.getDoc().setValue(editors.code.getDoc().getValue().split('\t').join(Array(editors.code.getOption('indentUnit') + 1).join(' '))) 
        /*if (appData.currentProject && appData.currentProject.lint_info && CodeEditor.curtab>=0){
            var curtab = CodeEditor.openTabs[CodeEditor.curtab]
            var info = appData.currentProject.lint_info
            if (curtab && curtab.file && info.hasOwnProperty(curtab.file)){
                var linfo = info[curtab.file]["C"]
                var crts = 0
                var todo = {}
                for(var i=0;i<linfo.length;i++){
                    if (linfo[i].hasOwnProperty("dots")){
                        if (linfo[i].type=="C0326"){
                            //exactly one comma
                            if (!todo.hasOwnProperty(linfo[i].line))
                                todo[linfo[i].line]=[]
                        }
                    }
                }
            }
        }*/
    }
    ,
    fill_opened_projects: function() {
        var container = $('#projectsOpened')
        container.empty();
        var added = {}
        for (var i=0;i<appData.openedProjectsList.length;i++){
            var proj = appData.openedProjectsList[i];
            console.log("FILLING WITH")
            console.log(proj)
            var li = '<li class="list-group-item"><span class="badge '+((appData.currentProject.id==proj.id) ? ("projSelected"):("projStyle"+proj.style))+'" ' 
                + ' aria-hidden="true">'
                if(proj.id!=appData.currentProject.id)
                    li+='<span class="fa fa-close" onclick="javascript:CodeEditor.closeAllTabsOf(\'' + proj.id + '\')"></span>'
                
                li+='<a href="javascript:Projects.select(\'' + proj.id + '\')">'+ proj.name+ '</a>'
                li+='</span></li>';
            container.append(li)
        }
        if (appData.openedProjectsList.length > 0) {
            $('#projectOpenedHeader').show();
        } else {
            $('#projectOpenedHeader').hide();
        }
    },
    closeAllTabsOf: function(pid) {
        var tabs=CodeEditor.openTabs
        var rem = []
        for(var i=0;i<tabs.length;i++){
            if(tabs[i].pid==pid){
                rem.push(i)
                CodeEditor.saveTab(i)
            }
        }
        rem.sort(function(a, b){return a-b});
        for (var i = rem.length -1; i >= 0; i--){
            if (CodeEditor.currentTab>=rem[i])
                CodeEditor.currentTab--;
            tabs.splice(rem[i],1);
        }

        if(CodeEditor.openTabs.length>0 && CodeEditor.currentTab<0)
            CodeEditor.currentTab=0


        CodeEditor.drawTabs()
    },
    refreshTabs:function(){
        var tabs = CodeEditor.openTabs
        for(var i=0;i<tabs.length;i++){
            if (tabs[i].pid==appData.currentProject.id){
                tabs[i].doc.setValue(window.atob(appData.currentProject.files[tabs[i].file]))
            }
        }
    },
    fill_project_from_tabs: function(proj){
        console.log("Filling from tabs")
        if (!proj) return
        var tabs = CodeEditor.openTabs
        for(var i=0;i<tabs.length;i++){
            if(tabs[i].pid==proj.id){
                proj.files[tabs[i].file]=window.btoa(tabs[i].doc.getValue())
            }
        }
        console.log(proj)
    },
    createTabDomElement: function(projectObj, fileName) {
        var template = $("#openFilsTabset").find(".tabFileOpen").first();
        var newElement = template.clone();
        newElement.attr("title", projectObj.name);
        newElement.find(".filename").text(fileName);
        newElement.data("projectId", projectObj.id);
        newElement.data("fileName", fileName);
        newElement.data("projName", projectObj.name);
        newElement.attr("id", "_tab" + projectObj.id + "-" + fileName.replace(/\./g, ""))
        newElement.unbind('click').click(function() {
            CodeEditor.editFileInEditor(projectObj, fileName)
        })
        newElement.css('display', 'inline-block');
        newElement.addClass('tabSelected');
        return newElement
    },
    refreshOpenedFileSelect: function() {
        if (CodeEditor.getNumbersOfOpenTabs() > appData.configuration.maxNumberOfOpenedTabs) {
            var overNumberSelectedTab = $("#openFilsTabset").find(".tabFileOpen").eq(-2);
            var ele = $("<li><a href='#'>" + overNumberSelectedTab.find(".filename").text() + "</a></li>");
            ele.data("projectId", overNumberSelectedTab.data("projectId"));
            ele.data("fileName", overNumberSelectedTab.data("fileName"));
            ele.click(function() {
                var projectObj = Projects.getProjectFromId($(this).data("projectId"));
                var fileName = $(this).data("fileName");
                CodeEditor.editFileInEditor(projectObj, fileName);
                $(this).remove();
            })
            $("#openedFileNotInToolbar").append(ele);
            overNumberSelectedTab.remove();
            $("#OpenedFileNotInToolbarButton").show()
        } else {
            var element = $("#openedFileNotInToolbar li").eq(1);
            if (element.length > 0) {
                var projectId = element.data("projectId");
                var fileName = element.data("fileName");
                var elementToInsert = CodeEditor.createTabDomElement(Projects.getProjectFromId(projectId), fileName);
                $("#openFilsTabset").append(elementToInsert);
            }
        }
    },
    clearErrors: function() {
            if (editors && editors.code)
                editors.code.clearGutter("py-marker");
    },
    makeMarker: function(txt,icon,_class){
        var marker = $('<div>', {
            'data-toggle': 'tooltip',
            'class': _class+' tooltip-bl',
            'title': txt,
            'data-delay': "0",
            'data-trigger': "hover focus click",
            'data-placement': 'right',
            html: '<span class="fa '+icon+'"></span>',
           
        })[0];
        return marker
    }
    ,
    makeErrorMarker: function(txt, warning) {
        var mcol;
        if (warning) mcol = "yellow"
        else mcol = "#C22"
        var marker = $('<div>', {
            'data-toggle': 'tooltip',
            'class': 'py-error-tag tooltip-bl',
            'title': txt,
            'data-delay': "0",
            'data-trigger': "hover focus click",
            'data-placement': 'right',
            html: '<span class="fa fa-close"></span>',
            css: {
                color: mcol
            }
        })[0];
        //marker.style.color = "#C22";
        //\marker.innerHTML = "&#9658;";
        return marker;
    },
    makeExceptionMarker: function(js) {
        var marker = $('<div>', {
            'class': 'py-exc-tag',
            'title': "Exception: click for traceback",
            html: '<span class="fa fa-close" onclick="'+js+'"></span>',
        })[0];
        //marker.style.color = "#C22";
        //\marker.innerHTML = "&#9658;";
        return marker;
    }
}