function toUTF8Array(str) {
    var utf8 = [];
    for (var i=0; i < str.length; i++) {
        var charcode = str.charCodeAt(i);
        if (charcode < 0x80) utf8.push(charcode);
        else if (charcode < 0x800) {
            utf8.push(0xc0 | (charcode >> 6), 
                      0x80 | (charcode & 0x3f));
        }
        else if (charcode < 0xd800 || charcode >= 0xe000) {
            utf8.push(0xe0 | (charcode >> 12), 
                      0x80 | ((charcode>>6) & 0x3f), 
                      0x80 | (charcode & 0x3f));
        }
        // surrogate pair
        else {
            i++;
            // UTF-16 encodes 0x10000-0x10FFFF by
            // subtracting 0x10000 and splitting the
            // 20 bits of 0x0-0xFFFFF into two halves
            charcode = 0x10000 + (((charcode & 0x3ff)<<10)
                      | (str.charCodeAt(i) & 0x3ff))
            utf8.push(0xf0 | (charcode >>18), 
                      0x80 | ((charcode>>12) & 0x3f), 
                      0x80 | ((charcode>>6) & 0x3f), 
                      0x80 | (charcode & 0x3f));
        }
    }
    return utf8;
}

var ViperClient = {
    socket: null,
    opened: false,
    queue: {},
    socket: null,
            nmsg: 0,
    cbks: {},
    acbk: {},
    init: function(port) {
       
        var me = this
        console.log("trying to open socket on port " + port)
        me.socket = Websock();
        me.socket.on("message", me.onMessage)
        me.socket.on("open", me.onOpen)
        me.socket.on("error", me.onError)
        me.socket.on("close", me.onClose)
        me.port = port
    },
    start: function(){
        ViperClient.socket.open("ws://127.0.0.1:" + ViperClient.port, ['binary']);
    },
    onMessage: function() {
        console.log("received")
        var data = ViperClient.socket.rQshiftStr(ViperClient.socket.rQlen())
        var msg = JSON.parse(data)
        console.log(msg)
        if (ViperClient.cbks.hasOwnProperty(msg.type)) {
            if(ViperClient.cbks[msg.type])
                ViperClient.cbks[msg.type](msg)
        } else {
            if (ViperClient.acbk.hasOwnProperty(msg.type+"_"+msg.id)) {
                ViperClient.acbk[msg.type+"_"+msg.id](msg)
                delete ViperClient.acbk[msg.type+"_"+msg.id]
            } else console.log("no registered callback")
        }
    },
    onOpen: function() {
        console.log("opened")
        ViperClient.opened = true
        ViperClient.cbks["open"]()
        ViperClient.send({"type": "hello"})
    },
    onClose: function() {
        console.log("closed")
        ViperClient.opened = false
        ViperClient.cbks["close"]()
    },
    onError: function(msg) {
       
        console.log("error")
        ViperClient.opened = false
        ViperClient.cbks["error"]()
    },
    on: function(msg, callback) {
        ViperClient.cbks[msg] = callback
    },
    send: function(msg, callback) {

        if (ViperClient.opened) {
            console.log("SENDING")
            console.log(msg)
            msg.id = ViperClient.nmsg++
            ViperClient.socket.send(toUTF8Array(JSON.stringify(msg)))
            if (callback)
                ViperClient.acbk[msg.type+"_"+msg.id] = callback
        }
    },
    answer: function(msg, data) {
        if (ViperClient.opened) {
            console.log("Answering")
            var res = {id: msg.id, type: msg.type, data: data}
            ViperClient.socket.send(toUTF8Array(JSON.stringify(res)))
            
        }
    }

}