import platform
import subprocess

class Node():
	nnodes = 0
	tree = {}

	@staticmethod
	def _reset():
		Node.nnodes=0
		Node.tree={}

	def __init__(self,name,level,cls,parent):
		self.name = name
		self.id = Node.nnodes
		Node.nnodes+=1
		self.level=level 
		self.cls=cls
		self.children = []
		self.attrs = {}
		self.ancestors = set()
		self.descendants = set()
		self.parent = parent
		Node.tree[self.id]=self
		if self.parent is not None:
			self.parent.children.append(self)
			self.ancestors.update(self.parent.ancestors)
			self.ancestors.add(self.parent.id)
			for ancestor in self.ancestors:
				Node.tree[ancestor].descendants.add(self.id)

	def has_vendor(self):
		return self.attrs.get("idVendor",False)

	def properties(self):
		return self.attrs
	def get_property(self,name):
		return self.attrs[name]
	def add_property(self,name,val):
		self.attrs[name]=val
		if name=="idVendor" or name=="idProduct":
			try:
				val=val[val.find("0x")+2:].upper()
				if len(val)<4:
					val = ("0"*(4-len(val)))+val
				self.attrs[name]=val
			except:
				pass

	def __str__(self):
		res=""
		res+="NODE["+str(self.id)+"]: "+self.name+" :: "+self.cls+"\n"
		res+="     parent: "+(self.parent.name if self.parent is not None else str(None))+"\n"
		res+="  ancestors: "+str(self.ancestors)+"\n"
		res+="descendants: "+str(self.descendants)+"\n"
		res+="   children: "+str(len(self.children))+"\n"
		res+="      attrs: "+str(len(self.attrs))+"\n"
		for k,v in self.attrs.items():
			res+="            "+str(k)+" = "+str(v)+"\n"
		res+="\n"
		return res


class MacUsb():
	def __init__(self):
		ver = platform.mac_ver()
		ver = ver[0]
		flds = ver.split(".")
		if int(flds[1])>=11:
			self.iousbstring = "IOUSBHostDevice"
		else:
			self.iousbstring = "IOUSBDevice"
		self.devices = {}
		self.nodes = {}
		self.text = ""

	def ioreg(self):
		try:
			#with open("/Users/giacomo/Viper/repo/ide/mac2.txt","r") as ff:
			#	text=ff.read()
			text = subprocess.check_output("ioreg -x -r -l -c "+self.iousbstring,universal_newlines=True,shell=True)
			if text==self.text:
				return
			self.text=text
			lines = text.split("\n")
			self.nodes = {}
			nnodes = 0
			parentstack = [None]
			Node._reset()
			for line in lines:
				nodepos = line.find("+-o")
				#print("check line",line)
				if nodepos>=0:
					clspos = line.find("<class")
					nodename = line[nodepos+3:clspos].strip()
					nodeclass = line[clspos+6:line.find(",",clspos)].strip()
					#nodelevel = line[0:nodepos].count("|")+ (0 if line.startswith("+") else 1)
					nodelevel = (nodepos+2)//3
					while parentstack[-1]!=None and parentstack[-1].level>=nodelevel:
						parentstack.pop()
					node = Node(nodename,nodelevel,nodeclass,parentstack[-1])
					parentstack.append(node)
				else:
					node = parentstack[-1]
					if node is None:
						continue
					eqpos = line.find('" =')
					if eqpos>=0:
						prpos = line[0:eqpos-1].find('"')
						prpname = line[prpos+1:eqpos]
						prpval =line[eqpos+3:].strip()
						node.add_property(prpname,prpval)
			self.nodes = Node.tree;
			#for k,node in self.nodes.items():
			#	print(str(node))
		
		except Exception as e:
			print(e)
			raise e

	def get_nodes(self):
		self.ioreg()
		return self.nodes


