import time
import copy
import json
import datetime
import uuid
import requests
import urllib.request

from ide.viperthreads import ViperThread

import os
import os.path
import sys

from viper.parse_rest.user import User
from viper.parse_rest.connection import SessionToken, register
from viper.parse_rest.core import *
from viper.parse_rest.datatypes import Function,Event

import ide.viper

import viper.viperconf as vconf
import viper.utils as vutils
import viper.toolbelt as tb
import viper.viperlogger

logger = viper.viperlogger.logger(__name__,"ide")



class ViperBackend(ViperThread):
    def __init__(self):
        super().__init__()
        self.db = ide.viper.getDB()
        self.user, self.password, self.email = self.db.get_user()
        self.bck_user = None
        self.logged = None
        self.logged_anonymously = None
        self.sent_info = None
        self.updated = None
        self.last_update_checked = 0
        self.remote_version=None
        self.logged_out = False
        self.gitkey = ""
        self.connected=False
    def start(self):
        #self.token = request.session.get('session_token')
        register(vconf.app_id, vconf.app_key)
        self._start(self.run,"Parse Backend")

    def check_connectivity(self):
        if vconf.platform.startswith("win"):
            response = os.system("ping 8.8.8.8 -n 1")
        else:
            response = os.system("ping -c 1 8.8.8.8")
        return response==0

    def run(self):
        login_retry=0
        login_last_time = 0
        login_info_last_time=0
        last_conn_check = 0
        self.connected = False
        while self.running:
            now = time.monotonic()
            if now-last_conn_check>60:
                cntct = self.check_connectivity()
                if cntct!=self.connected:
                    self.connected=cntct
                    ide.viper.getController().send_status()
                last_conn_check = now
            if self.connected:
                #logger.info("backend! %s %s",str(self.logged),str(self.user))
                if not self.logged and not self.user:
                    #user needs to signup, stay anonymous
                    #print("logA",self.logged_anonymously)
                    if not self.logged_anonymously:
                        if now-login_last_time>login_retry*15:
                        # 0 15 30 45 60....
                            login_last_time=now
                            res,out=self.login_anonymous()
                            if not res:
                                login_retry+=1
                                if login_retry>=40:
                                    login_retry=40
                            else:
                                login_retry=0
                    elif not self.sent_info:
                        #print("logA send")
                        self.login_info()
                elif not self.logged and self.user:
                    #try logging
                    if now-login_last_time>login_retry*15 and not self.logged_out:
                        # 0 15 30 45 60....
                        login_last_time=now
                        logger.info("Trying login")
                        ide.viper.say("REMOTE","Trying to login...")
                        res, out = self.login(self.user,self.password,self.email)
                        if not res:
                            if out=="ok":
                                ide.viper.say("REMOTE","Not logged, verify your email first")
                                login_retry=1
                            else:
                                ide.viper.say("REMOTE","Not logged, errors in the backend ("+str(out)+")")
                                login_retry+=1
                                if login_retry>=40:
                                    login_retry=40
                        else:
                            ide.viper.say("REMOTE","Logged!")
                            self.check_update()
                            login_retry=0
                elif self.logged and ( not self.sent_info or (now-login_info_last_time)>60*15):
                    logger.info("Trying login info")
                    self.login_info()
                    login_info_last_time=now
                    if self.remote_version:
                        if self.remote_version["version"]>vconf.get_full_viper_version():
                            ide.viper.say("REMOTE","There is a new update! v"+self.remote_version["version"]+" current is v"+vconf.get_full_viper_version())
                        ide.viper.getController().send_status()
                elif self.logged and not self.updated:
                    #self.synch_remote()
                    pass
                elif self.logged:
                    pass
                time.sleep(5)
            else:
                time.sleep(5)

    def create_user(self,user,password,email,job):
        if self.logged_anonymously:
            self.logout()
            self.bck_user=None
        try:
            ide.viper.say("REMOTE","Sign up user ["+user+"]...")
            u = User.signup(email, password,email=email,display_name=user,job=job,extra_headers={"X-Parse-Installation-Id":self.db.uid})
            self.db.store_user(user,password,email)
            self.user = user
            self.password=password
            self.email=email
            ide.viper.say("REMOTE","user ["+user+"] created, check your email to verify the account")
        except ParseError as pe:
            logger.error("creating Parse.User(%s,%s) got %s",user,email,str(pe))
            return (False,self.error_to_msg(pe))
        except Exception as e:
            logger.error("creating Parse.User(%s,%s) got %s",user,email,str(e))
            return (False,str(pe))
        return (True,"ok")

    def login_info(self):
        info = {
            "version":vconf.get_full_viper_version(),
            "system": vconf.get_short_machine_info(),
            "ip":vutils.get_public_ip()
        }
        res = self._call_cloud_code("login_info_ex",info=json.dumps(info))
        if res:
            self.sent_info = res["url"]
            self.gitkey = res["key"]
        else:
            self.sent_info = None
            self.gitkey=""
        if not info["ip"]:
            self.sent_info=None
        #print(self.sent_info)

    def error_to_msg(self,pe):
        s = pe.args[0].decode("utf-8")
        r = json.loads(s)
        if "error" in r:
            return r["error"]
        else:
            return s

    def reset_password(self,email):
        if self.logged:
            return (False,"You must be not logged!")
        if self.email and self.email!=email:
            return (False,"You must can only reset password for "+self.email)
        try:
            ret = User.request_password_reset(email)
            ide.viper.say("REMOTE","an email has been sent to "+email+" for password reset")
            return (True,ret)
        except ParseError as pe:
            logger.error("logging Parse.Reset(%s) got %s",email,str(pe))
            return (False,self.error_to_msg(pe))
        except Exception as e:
            logger.error("logging Parse.Reset(%s) got %s",email,str(e))
            return (False,str(e))


    def login(self,user,password,email=None):
        #u = User.signup(user, password,email=email)
        if self.logged_anonymously:
            self.logout()
            self.bck_user=None
        if self.email and email!=self.email:
            return (False, "You can only login as "+self.email)
        try:
            u = User.login(email,password,extra_headers={"X-Parse-Installation-Id":self.db.uid})
            self.bck_user = u
            self.logged = self.bck_user.emailVerified
            self.user = self.bck_user.display_name
            self.email = self.bck_user.email
            #trigger new login_info
            self.sent_info = None
            if self.logged:
                logger.info("Logged %s",self.user)
                self.db.store_user(self.user,password,self.bck_user.email)
                ide.viper.getController().send_status()
            else:
                logger.info("Not Logged %s",self.user)
                #User.logout()
            return (self.logged,"ok")
        except ParseError as pe:
            logger.error("logging Parse.User(%s) got %s",user,str(pe))
            return (False,self.error_to_msg(pe))
        except Exception as e:
            logger.error("logging Parse.User(%s) got %s",user,str(e))
            return (False,str(e))

    def login_anonymous(self):
        try:
            u = User.login_auth({
                "anonymous": {
                    "id":self.db.uid,
                    "username":self.db.uname,
                    "password":self.db.upass,
                    "email":"anonymous@viperize.it"
                }
                },extra_headers={"X-Parse-Installation-Id":self.db.uid})
            self.bck_user=u
            self.logged_anonymously = u
            return (self.logged_anonymously,"ok")
        except ParseError as pe:
            logger.error("logging anonymous Parse.User got %s",str(pe))
            return (False,self.error_to_msg(pe))
        except Exception as e:
            logger.error("logging anonymous Parse.User got %s",str(e))
            return (False,str(e))

    def logout(self):
        if self.logged:
            try:
                ide.viper.say("REMOTE","Logging out...")
                with SessionToken(self.bck_user.sessionToken):
                    User.logout()
                    self.bck_user=None
                    self.logged = None
                    self.logged_out = True
            except Exception as e:
                logger.error("logging   out Parse.User got %s",str(e))
                return False
            return True

    def can_i_git(self,proj):
        url = self.get_project_repo_url(proj)
        try:
            r= requests.get(url,timeout=10)
            if r.status_code==200:
                return True
            logger.error("Can I git? %s",str(r.status_code))
        except Exception as e:
            logger.error("Can I git? %s",str(e))
        return False

    def rate_package(self,pkg,rating):
        if not self.logged:
            return None
        return self._call_cloud_code("rate_package",pkg=pkg,rating=rating)

    def dload_package(self,pkg):
        if not self.logged:
            return None
        return self._call_cloud_code("download_package",pkg=pkg)


    def getUser(self):
        return self.user

    def getEmail(self):
        if self.logged:
            return self.bck_user.email
        return None

    def getUserId(self):
        if self.logged:
            return self.bck_user.objectId
        return None

    def getKey(self):
        if self.logged:
            return self.gitkey
        return None

    def get_credentials(self):
        if self.logged:
            return (self.getUserId(),self.getKey())
        return None

    def synch_remote(self):
        prjs = self.get_projects()
        if prjs!=None:
            for prj in prjs:
                ide.viper.getController().set_project_remote(prj)

    def synch(self,projects):
        pass

    def encrypt(self,data):
        pass

    def decrypt(self,data):
        pass

    def send_stats(self,thestats):
        logger.info("sending stats...")
        return self._call_cloud_code("stats",stats=self.prepare_stats(thestats))

    def sync_pkglist(self):
        pkglist = [(pkg.fullname,str(pkg.installed)) for pkg in tb.vpm.get_all_installed_packages()]
        self._call_cloud_code("sync_packages",pkglist=pkglist)

    def _call_cloud_code(self,name,**kw):
        if not self.logged and not self.logged_anonymously:
            return None
        with SessionToken(self.bck_user.sessionToken):
            func = Function(name)
            attempts=0
            while attempts<3:
                try:
                    res = func(**kw)
                    logger.info("calling Parse.%s ok with %s",name,str(res))
                    if "result" in res:
                        return res["result"]
                except ParseError as pe:
                    logger.error("calling Parse.%s got %s",name,str(pe))
                    res=None
                except Exception as e:
                    logger.error("calling Parse.%s got (non-parse) %s",name,str(e))
                    res=None
                finally:
                    attempts+=1
            return None

    def get_update(self):
        return self.remote_version

    def check_update(self):
        if not self.logged:
            return None
        if time.monotonic()-self.last_update_checked>15*60:
            self.last_update_checked=time.monotonic()
            res = self._call_cloud_code("check_update_ex",platform=vconf.get_full_platform(),version=vconf.get_full_viper_version(),build=vconf.get_viper_build())
            if "msg" in res:
                self.remote_version = res
            else:
                self.remote_version=None
            return self.remote_version
        elif self.remote_version:
            return self.remote_version
        return None


    def get_projects(self):
        if not self.logged:
            return None
        prjs = self._call_cloud_code("get_projects")
        if prjs:
            for prj in prjs:
                prj["remoteId"]=prj["objectId"]
                if "libinfo" in prj:
                    if prj["libinfo"]:
                        prj["namespace"]=prj["libinfo"]["namespace"]
                        prj["mod_name"]=prj["libinfo"]["mod_name"]
                        prj["keywords"]=prj["libinfo"]["keywords"]
                        prj["local"]=prj["libinfo"]["local"]
                    del prj["libinfo"]
        return prjs

    def new_project(self,project):
        if not self.logged:
            return None
        return self._call_cloud_code("new_project",project=self.prepare_project(project))

    def delete_project(self,project):
        if not self.logged:
            return None
        return self._call_cloud_code("delete_project",remoteId=project["remoteId"])

    def update_project(self,project):
        if not self.logged:
            return None
        return self._call_cloud_code("update_project",project=self.prepare_project(project))

    def get_project_repo_url(self,proj):
        if self.sent_info:
            url=self.sent_info
        else:
            url="https://cloud.viperize.it/"
        return url+self.getUserId()+"/"+self.getKey()+"/git/"+self.getUserId()+"_"+proj["remoteId"]

    def prepare_stats(self,stats):
        res = []
        for k,stat in stats.items():
            nst = {"type":stat["type"],"content":json.dumps(stat)}
            res.append(nst)
        return res

    def prepare_project(self,proj):
        res = copy.deepcopy(proj)
        # First remove passwords from repos: we don't want them on the cloud :)
        repos = []
        #print(res)
        if "repos" in res:
            for repo in res["repos"]:
                if "pass" in repo:
                    del repo["pass"]
                repos.append(json.dumps(repo))
        res["repos"]=repos
        # Second prepare files and tags
        files = []
        if "files" in res:
            for fl in res["files"]:
                files.append(fl)
        res["files"]=files

        tags=[]
        if "tags" in res:
            for tag in res["tags"]:
                tags.append(tag)
        res["tags"]=tags

        libinfo={}
        if "namespace" in res:
            libinfo["namespace"]=res["namespace"]
            libinfo["mod_name"]=res["mod_name"]
            libinfo["local"]=res["local"]
            libinfo["keywords"]=res["keywords"]
        res["libinfo"]=libinfo

        return {k:v for k,v in res.items() if k in set(["tags","name","desc","files","repos","remoteId","libinfo"])}

    def create_namespace(self,namespace):
        return self._call_cloud_code("add_namespace",namespace=namespace)

    def list_namespaces(self):
        return self._call_cloud_code("list_namespaces")

    def create_package(self,libname,libdesc,liburl,namespace,libpath,libtag,libkwrd=""):
        pkg = {
            "name":libname,
            "description":libdesc,
            "url":liburl,
            "namespace":namespace,
            "path":libpath,
            "tag":libtag,
            "keywords":libkwrd
        }
        return self._call_cloud_code("update_package",package=pkg)

    def list_packages(self):
        return self._call_cloud_code("list_packages")
