import sqlite3
import json
import os
import os.path
import threading
import ide.viper
import uuid
import time

# store of objects. Every object is serialized in json, must have a "uid" field used as key

class ViperDB():
    def __init__(self):
        self.db = sqlite3.connect(os.path.abspath(os.path.join(ide.viper.projdir,ide.viper.dbfile)),check_same_thread=False)
        self.db.execute("""CREATE TABLE IF NOT EXISTS projects(uid TEXT PRIMARY KEY, json TEXT)""")
        self.db.execute("""CREATE TABLE IF NOT EXISTS devices(uid TEXT PRIMARY KEY, json TEXT)""")
        self.db.execute("""CREATE TABLE IF NOT EXISTS user(email TEXT PRIMARY KEY, password TEXT, name TEXT)""")
        self.db.row_factory = sqlite3.Row
        self.dbs = sqlite3.connect(os.path.abspath(os.path.join(ide.viper.projdir,ide.viper.statfile)),check_same_thread=False) 
        self.dbs.execute("""CREATE TABLE  IF NOT EXISTS stats(uid INT PRIMARY KEY, json TEXT)""")
        self.dbs.row_factory = sqlite3.Row
        try:
            with open(os.path.abspath(os.path.join(ide.viper.projdir,ide.viper.db2file)),"r") as ff:
                self.conf = json.load(ff)
        except:
            self.conf={"theme":"light-dark","stats":True,"first-time":1}
        try:
            with open(os.path.abspath(os.path.join(ide.viper.confdir,"uid")),"r") as ff:
                jj = json.load(ff)
                self.uid = jj["uid"]
                self.uname = jj["uname"]
                self.upass = jj["upass"]
        except:
            self.uid=str(uuid.uuid4())
            self.uname=str(uuid.uuid4())
            self.upass=str(uuid.uuid4())
        if "tutorials" not in self.conf:
            self.conf["tutorials"]={}
        for x in ["new-project-logged","new-project-anon","first-project-logged","first-project-anon","browse-projects-logged","browse-projects-anon","board-plugged-anon","board-plugged-logged"]:
            if x not in self.conf["tutorials"]:
                self.conf["tutorials"][x]=0
        if "reminder-anon" not in self.conf["tutorials"]:
            self.conf["tutorials"]["reminder-anon"]=str(time.time()).split(".")[0]

        self.dblock = threading.Lock()
    def _store_things(self,things,table,db):
        itr = []
        for k,v in things.items():
            itr.append( (k,json.dumps(v) ))
        self.dblock.acquire()
        if itr:
            c= db.executemany("INSERT OR REPLACE INTO "+table+" VALUES(?,?)",itr)
            c.close()
            db.commit()
        self.dblock.release()
    def store_projects(self,projects):
        self._store_things(projects,"projects",self.db)
    def store_devices(self,devices):
        self._store_things(devices,"devices",self.db)
    def store_stats(self,stats):
        self._store_things(stats,"stats",self.dbs)
    def store_user(self,name,password,email):
        self.dblock.acquire()
        try:
            c=self.db.execute("INSERT OR REPLACE INTO user VALUES(?,?,?)",(email,password,name))
            c.close()
            self.db.commit()
        except Exception as e:
            print(e)
        finally:
            self.dblock.release()
        
    def _get_things(self,table,db,limit=0):
        res={}
        self.dblock.acquire()
        try:
            q = "SELECT * FROM "+table+" ORDER BY uid"
            if limit>0:
                q+=" LIMIT "+str(limit)
            c = db.execute(q)
            while True:
                r = c.fetchone()
                if r==None:
                    break
                res[r[0]]=json.loads(r[1])
            c.close()
        finally:
            self.dblock.release()
        return res
    def get_projects(self):
        return self._get_things("projects",self.db)
    def get_devices(self):
        return self._get_things("devices",self.db)
    def get_stats(self,count):
        return self._get_things("stats",self.dbs,count)
    def get_user(self):
        res = (None,None,None)
        self.dblock.acquire()
        try:
            q = "SELECT * FROM user WHERE 1"
            c = self.db.execute(q)
            r=c.fetchone()
            if r:
                res = (r[2],r[1],r[0])
            c.close()
        except Exception as e:
            print(e)
        finally:
            self.dblock.release()
        return res
    def get_max_stat_uid(self):
        q = "SELECT max(uid) FROM stats"
        res = 0
        self.dblock.acquire()
        try:
            c = self.dbs.execute(q)
            if c==None:
                res=0
            else:
                r = c.fetchone()
                if r[0]==None:
                    res=0
                else:
                    res = r[0]+1
            c.close()
        finally:
            self.dblock.release()
        return res

    def _remove_things(self,table,db,ids):
        ids = [(x,) for x in ids]
        q = "DELETE FROM "+table+" WHERE uid=?"
        self.dblock.acquire()
        try:
            c=db.executemany(q,ids)
            c.close()
            db.commit()
        finally:
            self.dblock.release()
    def remove_stats(self,ids):
        self._remove_things("stats",self.dbs,ids)
    def remove_project(self,uid):
        self._remove_things("projects",self.db,[uid])
    def remove_device(self,device):
        self._remove_things("devices",self.db,[device["uid"]])

    def shutdown(self):
        self.db.commit()
        self.db.close()
        self.dbs.commit()
        self.dbs.close()
        try:
            with open(os.path.abspath(os.path.join(ide.viper.projdir,ide.viper.db2file)),"w") as ff:
                self.conf.update({"first-time":0})
                json.dump(self.conf,ff)
                #print("SAVING DB",self.conf)
            with open(os.path.abspath(os.path.join(ide.viper.confdir,"uid")),"w") as ff:
                json.dump({"uid":self.uid,"uname":self.uname,"upass":self.upass},ff)
        except:
            pass
    def get_current_project(self):
        return self.conf.get("last_project",None)

    def get_current_theme(self):
        return self.conf["theme"]

    def get_tutorials(self):
        return self.conf["tutorials"]

    def set_tutorials(self,tutorials):
        self.conf["tutorials"]=tutorials

    def set_current_theme(self,theme):
        self.conf["theme"]=theme

    def set_stats_usage(self,stats):
        self.conf["stats"]=stats

    def can_send_stats(self):
        if "stats" in self.conf:
            return self.conf["stats"]
        return True

    def set_current_project(self,proj):
        #print("CURRENT PROJECT",proj["id"])
        self.conf["last_project"]=proj["id"]

    def is_first_time(self):
        return self.conf.get("first-time",1)
