import os
import os.path
import stat
import webbrowser
import sys
import threading
import time
import json
import traceback
import subprocess
import shutil
import requests

import ide.viper as viperide
from ide.viperserver import ViperLooper
from ide.viperthreads import ViperThread

import viper.viperlogger
import viper.viperconf as vconf


logger = viper.viperlogger.logger(__name__,"ide")


def del_rw(action, name, exc):
    os.chmod(name, stat.S_IWRITE)
    try:
        if os.path.isdir(name):
            os.rmdir(name)
        else:
            os.remove(name)
    except:
        pass



def run():
    HPORT,WPORT = viperide.getServers()

    weburl = "http://127.0.0.1:"+str(HPORT)+"/ide/ide/index.html"

    attempts = 5

    while attempts>0:
        logger.info("Checking WEB server")
        attempts-=1
        try:
            r = requests.get(weburl,timeout=10)
            break
        except Exception as e:
            logger.info(str(e))

    ##Open DB
    try:
        viperide.getDB()
    except Exception as e:
        logger.error("Can't open database! %s",str(e))
        ViperThread.gracefully_exit()
        sys.exit(1)

    ### Start browser
    logger.info("Launching Web Browser")
    opt = {
        "name": "Viper",
        "version": "0.0.1",
        "main": weburl,
        "node-remote":"127.0.0.1:"+str(HPORT),
        "single-instance":False,
        "window":{
            "toolbar":False,
            "icon":"viper.png",
            "min_width":800,
            "min_height":600,
            "position":"center"
        }
    }

    if vconf.platform=="linux":
        with open(os.path.join(vconf.envdirs["sys"],"browser","package.json"),"w") as ff:
            json.dump(opt,ff)
        subprocess.Popen(os.path.join(vconf.envdirs["sys"],"browser","nw"), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)
    elif vconf.platform=="mac":
        with open(os.path.join(vconf.envdirs["sys"],"browser","nwjs.app","Contents","Resources","app.nw","package.json"),"w") as ff:
            json.dump(opt,ff)
        subprocess.Popen(os.path.join(vconf.envdirs["sys"],"browser","nwjs.app","Contents","MacOS","nwjs"), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)
    else:
        with open(os.path.join(vconf.envdirs["sys"],"browser","package.json"),"w") as ff:
            json.dump(opt,ff)
        subprocess.Popen(os.path.join(vconf.envdirs["sys"],"browser","nw.exe"), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)

    #return
    #webbrowser.open(weburl,new=2)


    ### Start Controller
    viperide.getController()

    ### Start Hw Listener
    viperide.getHwListener()

    ##Start Stat listener
    viperide.getStatListener()

    ##Start Remote Backend
    viperide.getBackend()

    ##Start Main Loop
    looper = viperide.getLooper()
    looper.start()
    looper.join()

    viperide.db.shutdown()
    viperide.getBackend().logout()
    logger.info("Bye! %s",str(viperide.restart))
    ViperThread.gracefully_exit()

    if viperide.restart:
        if vconf.platform=="linux" or vconf.platform=="mac":
            print(viperide.restart)
            subprocess.Popen(viperide.restart, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)
        elif vconf.platform=="windows":
            #Need shell=True, for elevation -_-
            os.chdir(vconf.bootstrapdir)
            subprocess.Popen([viperide.restart], stdout=subprocess.DEVNULL,stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL,shell=True)
    elif viperide.destroy:
        #OMG!
        if vconf.platform=="windows":
            shutil.rmtree(os.path.abspath(vconf.envdirs["projects"]),onerror=del_rw)
        else:
            shutil.rmtree(os.path.abspath(vconf.envdirs["projects"]),ignore_errors=True)
