import viper.viperlogger
import viper.viperconf as vconf
import threading
import json
import queue
import time
import base64
import os, sys, time, select
import socketserver
import socket
import urllib.parse
from ide.wsserver import WebSocketRequestHandler,WebSocketServer
from http.server import HTTPServer, SimpleHTTPRequestHandler

from ide.viperthreads import ViperThread
#from ide.vipercontrol import ViperControl
#print(sys.path)
import ide.viper
import viper.toolbelt as tb

logger = viper.viperlogger.logger(__name__,"ide")
thq = queue.Queue()
rq = queue.Queue()
wq = queue.Queue()


class WSHandler(WebSocketRequestHandler):
    def new_websocket_client(self):
        thq.put(self)
        self.exit_ = False
        c_pend = 0
        rlist = [self.request]
        closed = False
        rdata = bytearray()

        try:
            while not closed and not self.exit_:
                wlist = []
                #print("rq:",rq.qsize(),"wq:",wq.qsize(),"c_pend:",c_pend,"rdata:",len(rdata))
                if (not wq.empty()) or c_pend: wlist.append(self.request)
                ins, outs, excepts = select.select(rlist, wlist, [], 1)
                if excepts: raise Exception("Socket exception")

                if self.request in outs:
                    # Send queued messages
                    tosend = []
                    try:
                        while len(tosend)<1024:
                            tosend.append(bytes(json.dumps(wq.get_nowait())+"\n","utf-8"))
                    except queue.Empty:
                        pass
                    c_pend = self.send_frames(tosend)
                    cqueue = []
                    tosend = []

                if self.request in ins:
                    # Receive client data, decode it, and put in rq
                    frames, closed = self.recv_frames()
                    for frame in frames:
                        #print("frame",frame.decode("utf-8"))
                        rdata.extend(frame)
                        try:
                            #print("decoding msg")
                            #print(rdata)
                            #print(str(rdata,"utf-8"))
                            msg = json.loads(str(rdata,"utf-8"))
                            #print(msg)
                            rq.put(msg)
                            rdata.clear()
                        except Exception as e:
                            print(ide.viper.pretty_exc(e))
                            pass
                    if closed:
                        logger.info("WS handler closing")
                        self.send_close()
                        self.request.close()
                        ViperThread.gracefully_exit(None,None)
        except Exception as e:
            print(e)
            raise e
        if not closed:
            #print("Closing!!")
            self.send_close()
        self.finish()



class ViperServer(ViperThread):
    def __init__(self,port):
        super().__init__()
        self.port = port
        self.sock= None
        self.nmsg=0

    def start(self):
        try:
            self.start_server = WebSocketServer(WSHandler, listen_port=self.port,run_once=True,verbose=True)
            self._start(self.run,"Websocket Server",self.shutdown)
            th = threading.Thread(None,self.run,"websockets")
            return True
        except Exception as e:
            print(e)
            return False
    def send(self,msg):
        try:
            msg["id"]=self.nmsg
            wq.put(msg)
            self.nmsg+=1
        except Exception as e:
            time.sleep(1)
            logger.info("send Exception %s",str(e))
    def answer(self,msg,data,text=""):
        try:
            res={}
            res["type"]=msg["type"]
            res["id"]=msg["id"]
            res["data"]=data
            res["msg"]=text
            wq.put(res)
        except Exception as e:
            time.sleep(1)
            logger.info("answer Exception %s",str(e))
    def recv(self):
        try:
            return rq.get(timeout=1)
        except queue.Empty:
            return None
        except Exception as e:
            time.sleep(1)
            logger.info("recv Exception %s",str(e))
    def run(self):
        logger.info("Starting WebSocket Server")
        try:
            self.start_server.start_server()
        except Exception as e:
            logger.info("oops, something weird in ws run: "+str(e))
        #logger.info("Killed WebSocket Server")
    def shutdown(self):
        global thq
        if self.start_server:
            self.start_server.shutdown()
        try:
            while True:
                th = thq.get_nowait()
                th.exit_=True
                logger.info("Killing open WebSocket Handlers")
        except queue.Empty:
            pass



from email.utils import formatdate
from time import mktime
import datetime

class RequestHandler(SimpleHTTPRequestHandler):
    def log_message( self, format, *args ):
        logger.info("Served "+format,*args)

    def end_headers(self):
        self.send_my_headers()
        SimpleHTTPRequestHandler.end_headers(self)

    def send_my_headers(self):
        self.send_header("Cache-Control", "no-cache, no-store, must-revalidate, max-age=0")
        self.send_header("Pragma", "no-cache")
        now = datetime.datetime.now()
        stamp = mktime(now.timetuple())+3600
        self.send_header("Last-Modified",formatdate(timeval     = stamp,localtime   = False,usegmt      = True))

    def do_GET(self):
        if self.path.endswith("viperconf.js"):
            self.send_response(200, 'OK')
            self.send_header('Content-type', 'text/javascript')
            self.end_headers()
            conf = {"port":ide.viper.W_PORT}
            conf["theme"]=ide.viper.getDB().get_current_theme()
            conf["firsttime"]=ide.viper.getDB().is_first_time()
            conf["tutorials"]=ide.viper.getDB().get_tutorials()
            conf["stats"]=ide.viper.getDB().can_send_stats()
            conf["type"]=vconf.app_type
            conf["version"]=vconf.get_full_viper_version()
            conf["build"]=vconf.get_viper_build()
            conf["startup"]=vconf.get_startup_version()
            self.wfile.write(bytes("var ViperConf="+json.dumps(conf),"utf-8"))
        # elif self.path.startswith("/ide/"):
        #     try:
        #         scheme,netloc,path,_,_,_ = urllib.parse.urlparse(self.path[5:])
        #         target = os.path.join(vconf.envdirs["ide"],path)
        #         print(target,path)
        #         with open(target,"rb") as ff:
        #             self.wfile.write(ff.read())
        #     except Exception as e:
        #         print(e)
        #         self.send_response(404,"")
        elif self.path.startswith("/vms/"):
            self.send_header('Content-type', 'application/octet-stream')
            self.send_header('Content-Disposition', 'attachment')
            self.send_response(200, 'OK')
            vmid = self.path[self.path.find("/vms/")+5:]
            #print(vmid)
            vms = ide.viper.getController().vms
            if vmid in vms:
                bin = base64.b64decode(bytes(vms[vmid]["bin"],"utf-8"))
                self.wfile.write(bin)
            else:
                self.wfile.write(bytes())
        elif self.path=="/docs":
            # generate help page
            try:
                print("generating help page")
                path = tb.generate_doc_page()
                with open(path,"rb") as ff:
                    self.wfile.write(ff.read())
            except Exception as e:
                print(str(e))
                self.send_response(404,"Can't find main documentation index!")
        elif self.path.startswith("/projects/"):
            if '?' in self.path:
                self.path = self.path.split('?')[0]
            try:
                target = os.path.join(vconf.envdirs["projects"],self.path[10:])
                print("opening--->",target)
                with open(target,"rb") as ff:
                    self.wfile.write(ff.read())
            except Exception as e:
                print(str(e))
                self.send_response(404,"Can't find project documentation!")
        elif self.path.startswith("/docs/"):
            if '?' in self.path:
                self.path = self.path.split('?')[0]
            flds = self.path.split("/")
            try:
                pkguid = flds[2]
                pkgvrs = flds[3]
                pkg = tb.vpm.get_package(by=pkguid,version=tb.vpm.to_version(pkgvrs))
                target = os.path.join(pkg.get_folders(pkg.installed)["docs"],*flds[4:])
                with open(target,"rb") as ff:
                    self.wfile.write(ff.read())
            except Exception as e:
                print(str(e))
                self.send_response(404,"Can't find such documentation!")
        elif self.path.startswith("/pinmap/"):
            flds = self.path.split("/")
            try:
                brdname = flds[2]
                info = tb.get_board_info(brdname)
                print(os.path.join(info["path"],"svg",brdname+".svg"))
                with open(os.path.join(info["path"],"svg",brdname+".svg"),"rb") as ff:
                    self.wfile.write(ff.read())
            except Exception as e:
                print(e)
                self.send_response(200,"NO pinmap")
        else:
            super().do_GET()


class MyTCPServer(socketserver.TCPServer):
    def set_vpthread(self,th):
        self.th = th
    def server_bind(self):
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.server_address)



class ViperHTTPD(ViperThread):
    def __init__(self):
        super().__init__()
        os.chdir(vconf.envdirs["core"])
    def start(self,port):
        try:
            self.httpd = MyTCPServer(("", port), RequestHandler)
            self.httpd.set_vpthread(self)
            self._start(self.httpd.serve_forever,"Http Server",self.httpd.shutdown)
            return True
        except Exception as e:
            print(e)
            return False


class ViperLooper(ViperThread):
    def __init__(self):
        super().__init__()
        self.client = ide.viper.getClient()
        self.hwlst = ide.viper.getHwListener()
        self.ctrl = ide.viper.getController()
        self.lastsentping = time.monotonic()
        self.lastrecvpong = self.lastsentping
        self.lastfeed = time.monotonic()
        self.has_client = False
    def start(self):
        self._start(self.run,"Main Loop")
    def run(self):
        self.ctrl.say("Hello Viper!")
        started = time.monotonic()
        while self.running:
            msg = self.client.recv()
            now = time.monotonic()
            if msg:
                try:
                    logger.info("handling message: %s",msg["type"])
                    msgtype = msg["type"]
                    if "data" in msg:
                        msgdata = msg["data"]
                    if msgtype=="hello":
                        self.ctrl.send_projects()
                        self.ctrl.send_devices()
                        self.ctrl.send_vms()
                        self.ctrl.send_examples()
                        self.ctrl.send_status()
                        self.ctrl.send_feed()
                        self.ctrl.send_vpm_check_updated()
                        self.has_client=True
                        self.lastsentping = time.monotonic()
                        self.lastrecvpong = self.lastsentping
                    elif msgtype=="quit":
                        #ViperThread.gracefully_exit()
                        break
                    elif msgtype=="ping":
                        self.lastrecvpong=now
                    elif msgtype=="conf":
                        ide.viper.getDB().set_current_theme(msg["data"]["theme"])
                        ide.viper.getDB().set_stats_usage(msg["data"]["stats"])
                        ide.viper.getDB().set_tutorials(msg["data"]["tutorials"])
                        vconf.save_startup_version(msg["data"]["startup"])
                    elif msgtype=="destroy_and_close":
                        ide.viper.destroy=True
                        break
                    elif msgtype=="do_update":
                        ret = ide.viper.getUpdater().do_update()
                        if ret:
                            #restart
                            ide.viper.restart=ret
                            self.client.answer(msg,"ok")
                            self.ctrl.dialog("Update!","""
                                All files have been downloaded successfully!<br>
                                The installer should now start, and it may ask permission to update your system.<br><br>
                                Another instance of the Viper IDE will be launched at the end of the update process; this IDE will be shutdown, you may need to close the browser tab manually.<br><br>
                                See you in a moment...
                                """)
                            time.sleep(5)
                            #ide.viper.getClient().send({"type":"quit","data":"ok"})
                            break
                        else:
                            self.client.answer(msg,"ko")
                            self.ctrl.notification("ooops! Something wrong during the update process!","vipererror")

                    else:
                        self.ctrl.dispatch(msg)
                except Exception as e:
                    traceback.print_exc()
                    logger.info("Exception while processing msg: %s",str(e))
            ## handle ping
            if self.has_client:
                if (now-self.lastsentping)>30:
                    logger.info("sending ping")
                    self.client.send({"type":"ping"})
                    self.lastsentping = now
                elif now-self.lastrecvpong>70:
                    logger.info("Not receiving pongs...quitting")
                    #ViperThread.gracefully_exit()
                    break
                elif now-self.lastfeed>60*30:
                    self.lastfeed=now
                    self.ctrl.send_feed()
                    self.ctrl.send_vpm_check_updated()
                ## handle devices
                self.ctrl.check_devices()
            else:
                if now-started>40:
                    #exit if no client connects in 40 seconds
                    self.running=False
