import ide.viper
from ide.viperthreads import ViperThread
import threading
import time
import datetime


import viper.viperlogger

logger = viper.viperlogger.logger(__name__,"ide")

class ViperStats(ViperThread):
    def __init__(self):
        super().__init__()
        self.db = ide.viper.getDB()
        self.statcnt = self.db.get_max_stat_uid()
        self.lock = threading.Lock()
        self.backend = ide.viper.getBackend()
    def newstat(self,stat):
        if stat:
            self.lock.acquire()
            stat["uid"]=self.statcnt
            self.db.store_stats({self.statcnt:stat})
            self.statcnt+=1
            self.lock.release()
    def start(self):
        self._start(self.run,"Stats Listener")
    def run(self):
        last_stat_try = 0 #TODO: change to now, don't want to start sending at login
        while self.running:
            #logger.info("staaaats")
            now=time.monotonic()
            if now-last_stat_try>60*ide.viper.statinterval:
                ##get a bunch of stats and send them
                logger.info("trying to send stats")
                stats = self.db.get_stats(20)
                last_stat_try=now
                while stats and self.running:
                    if self.backend.logged or self.backend.logged_anonymously:
                        logger.info("trying to send stats %s",str(stats))
                        res = self.backend.send_stats(stats)
                        if res:
                            self.db.remove_stats(stats.keys())
                            stats = self.db.get_stats(20)
                            time.sleep(5)
                        else:
                            break
            else:
                time.sleep(5)
    def send_stat(self,stype,stat):
        res = {}
        logger.info("Send stat %s %s",str(stype),str(self.db.can_send_stats()))
        if not self.db.can_send_stats():
            return res
        res["type"]=stype
        res["when"]= datetime.datetime.utcnow().isoformat()
        if "remoteId" in stat and stat["remoteId"]:
            res["remoteId"]=stat["remoteId"]
        if stype=="compile":
            if "exception" in stat:
                res["exception"]=stat["exception"]
            if "errors" in stat:
                res["errors"]=stat["errors"]
            if "info" in stat:
                res["stats"]=stat["info"]
            if "board" in stat:
                res["board"]=stat["board"]
        elif stype=="upload":
            if "exception" in stat:
                res["exception"]=stat["exception"]
            if "errors" in stat:
                res["error"]=stat["error"]
            if "vm" in stat and "device" in stat and "board" in stat:
                res["vm"]=stat["vm"]
                res["device"]=stat["device"]
                res["board"]=stat["board"]
        if len(res)<=2:
            res = {}
        logger.info("Before stat %s %s",str(stype),str(len(res)))
        if res:
            self.newstat(res)
            logger.info("New stat %s %s",str(stype),str(res))
        return res
