import threading
import viper.viperconf as vconf
import os
import os.path
import datetime
import urllib.request
import shutil
import tarfile
import json
import zlib
import time
import glob

import ide.viper

terminals=["xdg-terminal","x-terminal-emulator","gnome-terminal","konsole","xfce4-terminal","lxterminal","xterm","urxvt"]

class ViperUpdater():
    def __init__(self):
        self.client = ide.viper.getClient()
        self.backend = ide.viper.getBackend()

    def has_update(self):
        vrs = self.backend.get_update()
        if vrs and (vconf.get_viper_build()<vrs["build"]):
            return vrs
        return None

    def download_hook(self, blocks, blocksize, filesize):
        if blocks==0 or blocks%100==0 or blocks*blocksize==filesize:
            #ide.viper.say("UPDATE","Downloaded "+str(blocks*blocksize)+"/"+str(filesize))
            self.client.send({"type":"update_stats","data":{ "down":blocks*blocksize,"total":filesize}})

    def do_download(self,vrs):
        ide.viper.say("UPDATE","Downloading "+vrs["url"])
        if vconf.platform.startswith("win"):
            vfile = os.path.join(ide.viper.upddir,"ViperUpdater"+vrs["version"]+".exe")
        else:
            vfile = os.path.join(ide.viper.upddir,"ViperUpdater"+vrs["version"]+".tar.xz")
        if os.path.exists(vfile):
            os.remove(vfile)
        self.client.send({"type":"update_stats","data":{ "text":"Downloading update "+vrs["version"]}})
        urllib.request.urlretrieve(vrs["url"],vfile,reporthook=self.download_hook)
        #shutil.copy2(os.path.join(os.path.expanduser("~"),"git","ViperMac","ViperTestingUpdate.tar.xz"),vfile)
        self.client.send({"type":"update_stats","data":{ "down":1,"total":1}})
        prevcrc=0
        with open(vfile,"rb") as crcf:
            while True:
                buf = crcf.read(1024*1024)
                if not buf:
                    break
                prevcrc = zlib.crc32(buf,prevcrc)
        #print(hex(prevcrc),vrs["crc"])
        if hex(prevcrc)==vrs["crc"]:
            ide.viper.say("UPDATE","Download complete!")
            return vfile
        else:
            ide.viper.say("UPDATE","ooops, downloaded file seems corrupted! (bad crc)")
            return None
        
    def do_unpack(self,vfile,vrs):
        if vfile:
            zp = tarfile.open(vfile,"r:xz")
            ide.viper.say("UPDATE","Extracting archive "+str(vfile))
            self.client.send({"type":"update_stats","data":{ "text":"Extracting package..."}})
            
            #cleanup old updates...
            for item in glob.glob(os.path.join(ide.viper.upddir,"upd_*")):
                if not os.path.isdir(item):
                    continue
                shutil.rmtree(item,ignore_errors=True)

            thedir = os.path.join(ide.viper.upddir,"upd_"+str(int(time.time())))
            shutil.rmtree(thedir,ignore_errors=True)
            zp.extractall(thedir)
            zp.close()
            if os.path.exists(os.path.join(thedir,"ViperUpdater","updater.sh")):
                os.chmod(os.path.join(thedir,"ViperUpdater","updater.sh"),0o755)
            if os.path.exists(os.path.join(thedir,"ViperUpdater","updater_unix.sh")):
                os.chmod(os.path.join(thedir,"ViperUpdater","updater_unix.sh"),0o755)
            if os.path.exists(os.path.join(thedir,"ViperUpdater","delete.sh")):
                os.chmod(os.path.join(thedir,"ViperUpdater","delete.sh"),0o755)
               
        
            return os.path.join(thedir,"ViperUpdater")


    def do_update(self):
        vrs = self.has_update()
        if vrs:
            try:
                ide.viper.say("UPDATE","Starting!")
                #cleaning up previous updates
                shutil.rmtree(ide.viper.upddir,ignore_errors=True)
                os.makedirs(ide.viper.upddir,exist_ok=True)
                vfile = self.do_download(vrs)
                if vconf.platform.startswith("win") or not vfile:
                    return vfile

                #linux & mac
                thedir = self.do_unpack(vfile,vrs["version"])

                if vconf.platform.startswith("mac"):
                    appname = "Viper.app" if vconf.app_type=="release" else "ViperTesting.app"
                    return (os.path.join(thedir,"updater.sh"),vconf.instdir,thedir,appname)
                else:
                    appname = "Viper" if vconf.app_type=="release" else "ViperTesting"
                    aterm = None
                    for term in terminals:
                        if shutil.which(term):
                            aterm = term
                            break
                    if aterm:
                        if not shutil.which("bash"):
                            ide.viper.say("UPDATE","ooops, failed: can't find bash! Please install it and update again.")
                            return False
                        else:
                            return (aterm,"-e",os.path.join(thedir,"updater_unix.sh"),vconf.instdir,thedir,vconf.homedir,vrs["version"],appname)
                    else:
                        ide.viper.say("UPDATE","ooops, failed: can't find a suitable terminal emulator. One of "+str(terminals)+" is needed")
                        return False  
            except Exception as e:
                ide.viper.say("UPDATE","ooops, failed: "+str(e))
                return False
        else:
            ide.viper.say("UPDATE","nothing to update")
        return False

